#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Aggregate handlers for Placement API."""

from oslo_db import exception as db_exc
from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from oslo_utils import timeutils
import webob

from placement import errors
from placement import exception
from placement import microversion
from placement.objects import resource_provider as rp_obj
from placement.policies import aggregate as policies
from placement.schemas import aggregate as schema
from placement import util
from placement import wsgi_wrapper


_INCLUDE_GENERATION_VERSION = (1, 19)


def _send_aggregates(req, resource_provider, aggregate_uuids):
    want_version = req.environ[microversion.MICROVERSION_ENVIRON]
    response = req.response
    response.status = 200
    payload = _serialize_aggregates(aggregate_uuids)
    if want_version.matches(min_version=_INCLUDE_GENERATION_VERSION):
        payload['resource_provider_generation'] = resource_provider.generation
    response.body = encodeutils.to_utf8(
        jsonutils.dumps(payload))
    response.content_type = 'application/json'
    if want_version.matches((1, 15)):
        req.response.cache_control = 'no-cache'
        # We never get an aggregate itself, we get the list of aggregates
        # that are associated with a resource provider. We don't record the
        # time when that association was made and the time when an aggregate
        # uuid was created is not relevant, so here we punt and use utcnow.
        req.response.last_modified = timeutils.utcnow(with_timezone=True)
    return response


def _serialize_aggregates(aggregate_uuids):
    return {'aggregates': aggregate_uuids}


def _set_aggregates(resource_provider, aggregate_uuids,
                    increment_generation=False):
    """Set aggregates for the resource provider.

    If increment generation is true, the resource provider generation
    will be incremented if possible. If that fails (because something
    else incremented the generation in another thread), a
    ConcurrentUpdateDetected will be raised.
    """
    # NOTE(cdent): It's not clear what the DBDuplicateEntry handling
    # is doing here, set_aggregates already handles that, but I'm leaving
    # it here because it was already there.
    try:
        resource_provider.set_aggregates(
            aggregate_uuids, increment_generation=increment_generation)
    except exception.ConcurrentUpdateDetected as exc:
        raise webob.exc.HTTPConflict(
            'Update conflict: %(error)s' % {'error': exc},
            comment=errors.CONCURRENT_UPDATE)
    except db_exc.DBDuplicateEntry as exc:
        raise webob.exc.HTTPConflict(
            'Update conflict: %(error)s' % {'error': exc})


@wsgi_wrapper.PlacementWsgify
@util.check_accept('application/json')
@microversion.version_handler('1.1')
def get_aggregates(req):
    """GET a list of aggregates associated with a resource provider.

    If the resource provider does not exist return a 404.

    On success return a 200 with an application/json body containing a
    list of aggregate uuids.
    """
    context = req.environ['placement.context']
    context.can(policies.LIST)
    uuid = util.wsgi_path_item(req.environ, 'uuid')
    resource_provider = rp_obj.ResourceProvider.get_by_uuid(
        context, uuid)
    aggregate_uuids = resource_provider.get_aggregates()

    return _send_aggregates(req, resource_provider, aggregate_uuids)


@wsgi_wrapper.PlacementWsgify
@util.require_content('application/json')
@microversion.version_handler('1.1')
def set_aggregates(req):
    context = req.environ['placement.context']
    context.can(policies.UPDATE)
    want_version = req.environ[microversion.MICROVERSION_ENVIRON]
    consider_generation = want_version.matches(
        min_version=_INCLUDE_GENERATION_VERSION)
    put_schema = schema.PUT_AGGREGATES_SCHEMA_V1_1
    if consider_generation:
        put_schema = schema.PUT_AGGREGATES_SCHEMA_V1_19
    uuid = util.wsgi_path_item(req.environ, 'uuid')
    resource_provider = rp_obj.ResourceProvider.get_by_uuid(
        context, uuid)
    data = util.extract_json(req.body, put_schema)
    if consider_generation:
        # Check for generation conflict
        rp_gen = data['resource_provider_generation']
        if resource_provider.generation != rp_gen:
            raise webob.exc.HTTPConflict(
                "Resource provider's generation already changed. Please "
                "update the generation and try again.",
                comment=errors.CONCURRENT_UPDATE)
        aggregate_uuids = data['aggregates']
    else:
        aggregate_uuids = data
    _set_aggregates(resource_provider, aggregate_uuids,
                    increment_generation=consider_generation)

    return _send_aggregates(req, resource_provider, aggregate_uuids)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


DEFAULT_CONSUMER_MISSING_ID = '00000000-0000-0000-0000-000000000000'

placement_group = cfg.OptGroup(
    'placement',
    title='Placement Service Options',
    help="Configuration options for connecting to the placement API service")

placement_opts = [
    cfg.BoolOpt(
        'randomize_allocation_candidates',
        default=False,
        help="""
If True, when limiting allocation candidate results, the results will be
a random sampling of the full result set. The
[placement]max_allocation_candidates config might limit the size of the full
set used as the input of the sampling.

If False, allocation candidates are returned in a deterministic but undefined
order. That is, all things being equal, two requests for allocation candidates
will return the same results in the same order; but no guarantees are made as
to how that order is determined.
"""),
    cfg.StrOpt(
        'incomplete_consumer_project_id',
        default=DEFAULT_CONSUMER_MISSING_ID,
        help="""
Early API microversions (<1.8) allowed creating allocations and not specifying
a project or user identifier for the consumer. In cleaning up the data
modeling, we no longer allow missing project and user information. If an older
client makes an allocation, we'll use this in place of the information it
doesn't provide.
"""),
    cfg.StrOpt(
        'incomplete_consumer_user_id',
        default=DEFAULT_CONSUMER_MISSING_ID,
        help="""
Early API microversions (<1.8) allowed creating allocations and not specifying
a project or user identifier for the consumer. In cleaning up the data
modeling, we no longer allow missing project and user information. If an older
client makes an allocation, we'll use this in place of the information it
doesn't provide.
"""),
    cfg.IntOpt(
        'allocation_conflict_retry_count',
        default=10,
        help="""
The number of times to retry, server-side, writing allocations when there is
a resource provider generation conflict. Raising this value may be useful
when many concurrent allocations to the same resource provider are expected.
"""),
    cfg.IntOpt(
        'max_allocation_candidates',
        default=-1,
        help="""
The maximum number of allocation candidates placement generates for a single
request. This is a global limit to avoid excessive memory use and query
runtime. If set to -1 it means that the number of generated candidates are
only limited by the number and structure of the resource providers and the
content of the allocation_candidates query.

Note that the limit param of the allocation_candidates query is applied after
all the viable candidates are generated so that limit alone is not enough to
restrict the runtime or memory consumption of the query.

In a deployment with thousands of resource providers or if the deployment has
wide and symmetric provider trees, i.e. there are multiple children providers
under the same root having inventory from the same resource class
(e.g. in case of nova's mdev GPU or PCI in Placement features) we recommend
to tune this config option based on the memory available for the
placement service and the client timeout setting on the client side. A good
initial value could be around 100000.

In a deployment with wide and symmetric provider trees we also recommend to
change the [placement]allocation_candidates_generation_strategy to
breadth-first.
"""),
    cfg.StrOpt(
        'allocation_candidates_generation_strategy',
        default="depth-first",
        choices=("depth-first", "breadth-first"),
        help="""
Defines the order placement visits viable root providers during allocation
candidate generation:

* depth-first, generates all candidates from the first viable root provider
  before moving to the next.

* breadth-first, generates candidates from viable roots in a round-robin
  fashion, creating one candidate from each viable root before creating the
  second candidate from the first root.

If the deployment has wide and symmetric provider trees, i.e. there are
multiple children providers under the same root having inventory from the same
resource class (e.g. in case of nova's mdev GPU or PCI in Placement features)
then the depth-first strategy with a max_allocation_candidates
limit might produce candidates from a limited set of root providers. On the
other hand breadth-first strategy will ensure that the candidates are returned
from all viable roots in a balanced way.

Both strategies produce the candidates in the API response in an undefined but
deterministic order. That is, all things being equal, two requests for
allocation candidates will return the same results in the same order; but no
guarantees are made as to how that order is determined.
"""),
]


# Duplicate log_options from oslo_service so that we don't have to import
# that package into placement.
# NOTE(cdent): Doing so ends up requiring eventlet and other unnecessary
# packages for just this one setting.
service_opts = [
    cfg.BoolOpt('log_options',
                default=True,
                help='Enables or disables logging values of all registered '
                     'options when starting a service (at DEBUG level).'),
]


def register_opts(conf):
    conf.register_group(placement_group)
    conf.register_opts(placement_opts, group=placement_group)
    conf.register_opts(service_opts)


def list_opts():
    return {placement_group.name: placement_opts}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_db.sqlalchemy import models
from oslo_log import log as logging
from sqlalchemy import Column
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import orm
from sqlalchemy import schema
from sqlalchemy import String
from sqlalchemy import Unicode

LOG = logging.getLogger(__name__)


class _Base(models.ModelBase, models.TimestampMixin):
    pass


BASE = orm.declarative_base(cls=_Base)


class ResourceClass(BASE):
    """Represents the type of resource for an inventory or allocation."""
    __tablename__ = 'resource_classes'
    __table_args__ = (
        schema.UniqueConstraint("name", name="uniq_resource_classes0name"),
    )

    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(String(255), nullable=False)


class ResourceProvider(BASE):
    """Represents a mapping to a providers of resources."""

    __tablename__ = "resource_providers"
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_resource_providers0uuid'),
        Index('resource_providers_root_provider_id_idx',
              'root_provider_id'),
        Index('resource_providers_parent_provider_id_idx',
              'parent_provider_id'),
        schema.UniqueConstraint('name', name='uniq_resource_providers0name')
    )

    id = Column(Integer, primary_key=True, nullable=False)
    uuid = Column(String(36), nullable=False)
    name = Column(Unicode(200), nullable=True)
    generation = Column(Integer, default=0)
    # Represents the root of the "tree" that the provider belongs to
    root_provider_id = Column(
        Integer, ForeignKey('resource_providers.id'), nullable=True)
    # The immediate parent provider of this provider, or NULL if there is no
    # parent. If parent_provider_id == NULL then root_provider_id == id
    parent_provider_id = Column(
        Integer, ForeignKey('resource_providers.id'), nullable=True)


class Inventory(BASE):
    """Represents a quantity of available resource."""

    __tablename__ = "inventories"
    __table_args__ = (
        Index('inventories_resource_class_id_idx',
              'resource_class_id'),
        schema.UniqueConstraint(
            'resource_provider_id', 'resource_class_id',
            name='uniq_inventories0resource_provider_resource_class')
    )

    id = Column(Integer, primary_key=True, nullable=False)
    resource_provider_id = Column(Integer, nullable=False)
    resource_class_id = Column(Integer, nullable=False)
    total = Column(Integer, nullable=False)
    reserved = Column(Integer, nullable=False)
    min_unit = Column(Integer, nullable=False)
    max_unit = Column(Integer, nullable=False)
    step_size = Column(Integer, nullable=False)
    allocation_ratio = Column(Float, nullable=False)
    resource_provider = orm.relationship(
        "ResourceProvider",
        primaryjoin=('Inventory.resource_provider_id == '
                     'ResourceProvider.id'),
        foreign_keys=resource_provider_id)


class Allocation(BASE):
    """A use of inventory."""

    __tablename__ = "allocations"
    __table_args__ = (
        Index('allocations_resource_provider_class_used_idx',
              'resource_provider_id', 'resource_class_id',
              'used'),
        Index('allocations_resource_class_id_idx',
              'resource_class_id'),
        Index('allocations_consumer_id_idx', 'consumer_id')
    )

    id = Column(Integer, primary_key=True, nullable=False)
    resource_provider_id = Column(Integer, nullable=False)
    consumer_id = Column(String(36), nullable=False)
    resource_class_id = Column(Integer, nullable=False)
    used = Column(Integer, nullable=False)
    resource_provider = orm.relationship(
        "ResourceProvider",
        primaryjoin=('Allocation.resource_provider_id == '
                     'ResourceProvider.id'),
        foreign_keys=resource_provider_id)


class ResourceProviderAggregate(BASE):
    """Associate a resource provider with an aggregate."""

    __tablename__ = 'resource_provider_aggregates'
    __table_args__ = (
        Index('resource_provider_aggregates_aggregate_id_idx',
              'aggregate_id'),
    )

    resource_provider_id = Column(Integer, primary_key=True, nullable=False)
    aggregate_id = Column(Integer, primary_key=True, nullable=False)


class PlacementAggregate(BASE):
    """A grouping of resource providers."""
    __tablename__ = 'placement_aggregates'
    __table_args__ = (
        schema.UniqueConstraint("uuid", name="uniq_placement_aggregates0uuid"),
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(String(36))


class Trait(BASE):
    """Represents a trait."""

    __tablename__ = "traits"
    __table_args__ = (
        schema.UniqueConstraint('name', name='uniq_traits0name'),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    name = Column(Unicode(255), nullable=False)


class ResourceProviderTrait(BASE):
    """Represents the relationship between traits and resource provider"""

    __tablename__ = "resource_provider_traits"
    __table_args__ = (
        Index('resource_provider_traits_resource_provider_trait_idx',
              'resource_provider_id', 'trait_id'),
    )

    trait_id = Column(Integer, ForeignKey('traits.id'), primary_key=True,
                      nullable=False)
    resource_provider_id = Column(Integer,
                                  ForeignKey('resource_providers.id'),
                                  primary_key=True,
                                  nullable=False)


class Project(BASE):
    """The project is the Keystone project."""

    __tablename__ = 'projects'
    __table_args__ = (
        schema.UniqueConstraint(
            'external_id',
            name='uniq_projects0external_id',
        ),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    external_id = Column(String(255), nullable=False)


class User(BASE):
    """The user is the Keystone user."""

    __tablename__ = 'users'
    __table_args__ = (
        schema.UniqueConstraint(
            'external_id',
            name='uniq_users0external_id',
        ),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    external_id = Column(String(255), nullable=False)


class Consumer(BASE):
    """Represents a resource consumer."""

    __tablename__ = 'consumers'
    __table_args__ = (
        Index('consumers_project_id_uuid_idx', 'project_id', 'uuid'),
        Index('consumers_project_id_user_id_uuid_idx', 'project_id', 'user_id',
              'uuid'),
        Index('consumers_consumer_type_id_idx', 'consumer_type_id'),
        schema.UniqueConstraint('uuid', name='uniq_consumers0uuid'),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    uuid = Column(String(36), nullable=False)
    project_id = Column(Integer, nullable=False)
    user_id = Column(Integer, nullable=False)
    generation = Column(Integer, nullable=False, server_default="0", default=0)
    consumer_type_id = Column(
        Integer, ForeignKey('consumer_types.id'), nullable=True)


class ConsumerType(BASE):
    """Represents a consumer's type."""

    __tablename__ = 'consumer_types'
    __table_args__ = (
        schema.UniqueConstraint('name', name='uniq_consumer_types0name'),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    name = Column(Unicode(255), nullable=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections

import sqlalchemy as sa

from placement.db.sqlalchemy import models
from placement import db_api
from placement import exception
from placement.objects import consumer_type as ct_obj

_CONSUMER_TYPE_TBL = models.ConsumerType.__table__
_RC_TBL = models.ResourceClass.__table__
_TRAIT_TBL = models.Trait.__table__


class _AttributeCache(object):
    """A cache of integer and string lookup values for string-based attributes.

    Subclasses must define `_table` and `_not_found` members describing the
    database table which is the authoritative source of data and the exception
    raised if data for an attribute is not found, respectively.

    The cache is required to be correct for the extent of any individual API
    request and be used only for those entities where any change to the
    underlying data is only making that change and will have no subsequent
    queries into the cache. For example, when we add a new resource class we
    do not then list all the resource classes from within the same session.

    Despite that requirement, any time an entity associated with a cache is
    created, updated, or deleted `clear()` should be called on the cache.
    """
    _table = None
    _not_found = None

    # The cache internally stores either sqlalchemy Row objects or
    # Attrs namedtuples but Row is compatible with namedtuple interface too.
    Attrs = collections.namedtuple(
        "Attrs", ["id", "name", "updated_at", "created_at"]
    )

    def __init__(self, ctx):
        """Initialize the cache of resource class identifiers.

        :param ctx: `placement.context.RequestContext` from which we can grab a
                    `SQLAlchemy.Connection` object to use for any DB lookups.
        """
        # Prevent this class being created directly, relevant during
        # development.
        assert self._table is not None, "_table must be defined"
        assert self._not_found is not None, "_not_found must be defined"
        self._ctx = ctx
        self.clear()

    def clear(self):
        self._id_cache = {}
        self._str_cache = {}
        self._all_cache = {}

    def id_from_string(self, attr_str):
        """Given a string representation of an attribute -- e.g. "DISK_GB"
        or "CUSTOM_IRON_SILVER" -- return the integer code for the attribute
        by doing a DB lookup into the appropriate table; however, the results
        of these DB lookups are cached since the lookups are so frequent.

        :param attr_str: The string representation of the attribute to look up
                         a numeric identifier for.
        :returns Integer identifier for the attribute.
        :raises An instance of the subclass' _not_found exception if attribute
                cannot be found in the DB.
        """
        attr_id = self._id_cache.get(attr_str)
        if attr_id is not None:
            return attr_id

        # Otherwise, check the database table
        self._refresh_from_db(self._ctx)
        if attr_str in self._id_cache:
            return self._id_cache[attr_str]
        raise self._not_found(name=attr_str)

    def all_from_string(self, attr_str):
        """Given a string representation of an attribute -- e.g. "DISK_GB"
        or "CUSTOM_IRON_SILVER" -- return all the attribute info.

        :param attr_str: The string representation of the attribute for which
                         to look up the object.
        :returns: namedtuple representing the attribute fields, if the
                  attribute was found in the appropriate database table.
        :raises An instance of the subclass' _not_found exception if attr_str
                cannot be found in the DB.
        """
        attrs = self._all_cache.get(attr_str)
        if attrs is not None:
            return attrs

        # Otherwise, check the database table
        self._refresh_from_db(self._ctx)
        if attr_str in self._all_cache:
            return self._all_cache[attr_str]
        raise self._not_found(name=attr_str)

    def string_from_id(self, attr_id):
        """The reverse of the id_from_string() method. Given a supplied numeric
        identifier for an attribute, we look up the corresponding string
        representation, via a DB lookup. The results of these DB lookups are
        cached since the lookups are so frequent.

        :param attr_id: The numeric representation of the attribute to look
                        up a string identifier for.
        :returns: String identifier for the attribute.
        :raises An instances of the subclass' _not_found exception if attr_id
                cannot be found in the DB.
        """
        attr_str = self._str_cache.get(attr_id)
        if attr_str is not None:
            return attr_str

        # Otherwise, check the database table
        self._refresh_from_db(self._ctx)
        if attr_id in self._str_cache:
            return self._str_cache[attr_id]
        raise self._not_found(name=attr_id)

    def get_all(self):
        """Return an iterator of all the resources in the cache with all their
        attributes as a namedtuple.

        In Python3 the return value is a generator.
        """
        if not self._all_cache:
            self._refresh_from_db(self._ctx)
        return self._all_cache.values()

    @db_api.placement_context_manager.reader
    def _refresh_from_db(self, ctx):
        """Grabs all resource classes or traits from the respective DB table
        and populates the supplied cache object's internal integer and string
        identifier dicts.

        :param ctx: RequestContext with the the database session.
        """
        table = self._table
        sel = sa.select(
            table.c.id,
            table.c.name,
            table.c.updated_at,
            table.c.created_at,
        )
        res = ctx.session.execute(sel).fetchall()
        self._id_cache = {r[1]: r[0] for r in res}
        self._str_cache = {r[0]: r[1] for r in res}
        # Note that r is Row object that is compatible with the namedtuple
        # interface of the cache
        self._all_cache = {r[1]: r for r in res}

    def _add_attribute(self, attr_id, name, created_at, updated_at):
        """Use this to add values to the cache that are not coming from the
        database, like defaults.
        """
        self._id_cache[name] = attr_id
        self._str_cache[attr_id] = name
        attrs = self.Attrs(attr_id, name, updated_at, created_at)
        self._all_cache[name] = attrs


class ConsumerTypeCache(_AttributeCache):
    """An _AttributeCache for consumer types."""

    _table = _CONSUMER_TYPE_TBL
    _not_found = exception.ConsumerTypeNotFound

    @db_api.placement_context_manager.reader
    def _refresh_from_db(self, ctx):
        super(ConsumerTypeCache, self)._refresh_from_db(ctx)
        # The consumer_type_id is nullable and records with a NULL (None)
        # consumer_type_id are considered as 'unknown'. Also the 'unknown'
        # consumer_type is not created in the database so we need to manually
        # populate it in the cache here.
        self._add_attribute(
            attr_id=None,
            name=ct_obj.NULL_CONSUMER_TYPE_ALIAS,
            # should we synthesize some dates in the past instead?
            created_at=None,
            updated_at=None,
        )


class ResourceClassCache(_AttributeCache):
    """An _AttributeCache for resource classes."""

    _table = _RC_TBL
    _not_found = exception.ResourceClassNotFound


class TraitCache(_AttributeCache):
    """An _AttributeCache for traits."""

    _table = _TRAIT_TBL
    _not_found = exception.TraitNotFound



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import sys
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslotest import output
import testtools

from placement.cmd import manage
from placement import conf
from placement.tests.unit import base


class TestCommandParsers(testtools.TestCase):

    def setUp(self):
        super(TestCommandParsers, self).setUp()
        self.conf = cfg.ConfigOpts()
        conf_fixture = config_fixture.Config(self.conf)
        self.useFixture(conf_fixture)
        conf.register_opts(conf_fixture.conf)
        # Quiet output from argparse (used within oslo_config).
        # If you are debugging, commenting this out might be useful.
        self.output = self.useFixture(
            output.CaptureOutput(do_stderr=True, do_stdout=True))
        # We don't use a database, but we need to set the opt as
        # it's required for a valid config.
        conf_fixture.config(group="placement_database", connection='sqlite://')
        command_opts = manage.setup_commands(conf_fixture)
        # Command line opts must be registered on the conf_fixture, otherwise
        # they carry over globally.
        conf_fixture.register_cli_opts(command_opts)

    def test_commands_associated(self):
        """Test that commands get parsed as desired.

        This leaves out --version, which is built into oslo.config's handling.
        """
        for command, args in [
                ('db_version', ['db', 'version']),
                ('db_sync', ['db', 'sync']),
                ('db_stamp', ['db', 'stamp', 'b4ed3a175331']),
                ('db_online_data_migrations',
                 ['db', 'online_data_migrations'])]:
            with mock.patch('placement.cmd.manage.DbCommands.' +
                            command) as mock_command:
                self.conf(args, default_config_files=[])
                self.conf.command.func()
                mock_command.assert_called_once_with()

    def test_non_command(self):
        """A non-existent command should fail."""
        self.assertRaises(SystemExit,
                          self.conf, ['pony'], default_config_files=[])

    def test_empty_command(self):
        """An empty command should create no func."""
        def parse_conf():
            self.conf([], default_config_files=[])

        def get_func():
            return self.conf.command.func

        parse_conf()
        self.assertRaises(cfg.NoSuchOptError, get_func)

    def test_too_many_args(self):
        self.assertRaises(SystemExit,
                          self.conf, ['version', '5'], default_config_files=[])
        self.output.stderr.seek(0)
        if sys.version_info >= (3, 12, 8):
            message = "choose from db"
        else:
            message = "choose from 'db'"
        self.assertIn(message, self.output.stderr.read())

    def test_help_message(self):
        """Test that help output for sub commands shows right commands."""
        self.conf(['db'], default_config_files=[])
        self.conf.command.func()

        self.output.stdout.seek(0)
        self.output.stderr.seek(0)

        self.assertIn('{sync,version,stamp,online_data_migrations}',
                      self.output.stdout.read())


class TestDBCommands(base.ContextTestCase):

    def setUp(self):
        super(TestDBCommands, self).setUp()
        self.conf = cfg.ConfigOpts()
        conf_fixture = config_fixture.Config(self.conf)
        self.useFixture(conf_fixture)
        conf.register_opts(conf_fixture.conf)
        conf_fixture.config(group="placement_database", connection='sqlite://')
        command_opts = manage.setup_commands(conf_fixture)
        conf_fixture.register_cli_opts(command_opts)
        self.output = self.useFixture(
            output.CaptureOutput(do_stderr=True, do_stdout=True))

    def _command_setup(self, max_count=None):
        command_list = ["db", "online_data_migrations"]
        if max_count is not None:
            command_list.extend(["--max-count", str(max_count)])
        self.conf(command_list,
                  project='placement',
                  default_config_files=None)
        return manage.DbCommands(self.conf)

    def test_online_migrations(self):
        # Mock two online migrations
        mock_mig1 = mock.MagicMock(__name__="mock_mig_1")
        mock_mig2 = mock.MagicMock(__name__="mock_mig_2")
        mock_mig1.side_effect = [(10, 10), (0, 0)]
        mock_mig2.side_effect = [(15, 15), (0, 0)]
        mock_migrations = (mock_mig1, mock_mig2)

        with mock.patch('placement.cmd.manage.online_migrations',
                        new=mock_migrations):
            commands = self._command_setup()
            commands.db_online_data_migrations()
            expected = '''\
Running batches of 50 until complete
10 rows matched query mock_mig_1, 10 migrated
15 rows matched query mock_mig_2, 15 migrated
+------------+-------------+-----------+
| Migration  | Total Found | Completed |
+------------+-------------+-----------+
| mock_mig_1 |      10     |     10    |
| mock_mig_2 |      15     |     15    |
+------------+-------------+-----------+
'''
            self.output.stdout.seek(0)
            self.assertEqual(expected, self.output.stdout.read())

    def test_online_migrations_error(self):
        good_remaining = [50]

        def good_migration(context, count):
            found = good_remaining[0]
            done = min(found, count)
            good_remaining[0] -= done
            return found, done

        bad_migration = mock.MagicMock()
        bad_migration.side_effect = Exception("Mock Exception")
        bad_migration.__name__ = 'bad'

        mock_migrations = (bad_migration, good_migration)

        with mock.patch('placement.cmd.manage.online_migrations',
                        new=mock_migrations):

            # bad_migration raises an exception, but it could be because
            # good_migration had not completed yet. We should get 1 in this
            # case, because some work was done, and the command should be
            # reiterated.
            commands = self._command_setup(max_count=50)
            self.assertEqual(1, commands.db_online_data_migrations())

            # When running this for the second time, there's no work left for
            # good_migration to do, but bad_migration still fails - should
            # get 2 this time.
            self.assertEqual(2, commands.db_online_data_migrations())

            # When --max-count is not used, we should get 2 if all possible
            # migrations completed but some raise exceptions
            commands = self._command_setup()
            good_remaining = [125]
            self.assertEqual(2, commands.db_online_data_migrations())

    def test_online_migrations_bad_max(self):
        commands = self._command_setup(max_count=-2)
        self.assertEqual(127, commands.db_online_data_migrations())

        commands = self._command_setup(max_count="a")
        self.assertEqual(127, commands.db_online_data_migrations())

        commands = self._command_setup(max_count=0)
        self.assertEqual(127, commands.db_online_data_migrations())

    def test_online_migrations_no_max(self):
        with mock.patch('placement.cmd.manage.DbCommands.'
                        '_run_online_migration') as rm:
            rm.return_value = {}, False
            commands = self._command_setup()
            self.assertEqual(0, commands.db_online_data_migrations())

    def test_online_migrations_finished(self):
        with mock.patch('placement.cmd.manage.DbCommands.'
                        '_run_online_migration') as rm:
            rm.return_value = {}, False
            commands = self._command_setup(max_count=5)
            self.assertEqual(0, commands.db_online_data_migrations())

    def test_online_migrations_not_finished(self):
        with mock.patch('placement.cmd.manage.DbCommands.'
                        '_run_online_migration') as rm:
            rm.return_value = {'mig': (10, 5)}, False
            commands = self._command_setup(max_count=5)
            self.assertEqual(1, commands.db_online_data_migrations())
